/* StatServ routines

   OperStats - Copyright (c) 2000-2001 Darcy Grexton
   Contact: cygnus@uniserve.ca, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/operstats.h"
#include "../inc/ss-help.h"

TLD *tldlist = NULL;
Boolean_T tldlookup = FALSE;

static void do_help (const char *source);
static void do_stats (const char *source);
static void do_map (const char *source);
static void do_servers (const char *source);
static void do_tld (const char *source);
static void do_reset (const char *source);
static void do_flush (const char *source);
static void do_auth (const char *source);
static void do_settings (const char *source);
static void do_deauth (const char *source);
static void do_shutdown (const char *source);
static void do_raw (const char *source);
static void do_inject (const char *source);
static void do_debug (const char *source);
static void do_update (const char *source);
static void do_rehash (const char *source);
static void do_settings (const char *source);
static void do_sendpings (const char *source);

/* Main StatServ routine. */
void statserv (const char *source, char *buf)  
{
    char *cmd, *s;
    char orig[BUFSIZE];

    /* Make a copy of the original message. */ 
    strscpy (orig, buf, sizeof (orig));   

    cmd = strtok (buf, " ");

    if (!cmd)
	return;
    
    if (!stricmp (cmd, "\1PING"))
    {
	if (!(s = strtok (NULL, "")))
	    s = "0";
	strip (s);
	ctcpreply (s_StatServ, source, "PING %s", s);
    }
    else if (!stricmp (cmd, "\1VERSION\1"))
	ctcpreply (s_StatServ, source, "VERSION OperStats v%s", version);

    /* A CTCP we don't recognize.. Ignore it. */
    else if (*cmd == '\1')
	return;

    else
    {
	Hash *command, hash_table[] =
	{
	    {"HELP",		H_NONE,		do_help},
	    {"STATS",		H_NONE,		do_stats},
	    {"TLD",		H_NONE,		do_tld},
	    {"MAP",		H_IRCOP,	do_map},
	    {"SERVERS",		H_IRCOP,	do_servers},
	    {"AUTH",		H_IRCOP,	do_auth},
	    {"DEAUTH",		H_SRA,		do_deauth},
	    {"FLUSH",		H_SRA,		do_flush},
	    {"RESET",		H_SRA,		do_reset},
	    {"SHUTDOWN",	H_SRA,		do_shutdown},
	    {"RAW",		H_SRA,		do_raw},
	    {"INJECT",		H_SRA,		do_inject},
	    {"DEBUG",		H_SRA,		do_debug},
	    {"UPDATE",		H_SRA,		do_update},
	    {"REHASH",		H_SRA,		do_rehash},
	    {"SENDPINGS",	H_SRA,		do_sendpings},
	    {"SETTINGS",	H_SRA,		do_settings},
	    {NULL}
	};
            
	if (command = get_hash (s_StatServ, source, strupper (cmd), hash_table))
	{
	    /* These are only available if OperServ is OFF. */
	    if ((command->process == do_auth ||
		 command->process == do_deauth ||
		 command->process == do_shutdown ||
		 command->process == do_raw ||
		 command->process == do_inject ||
		 command->process == do_debug ||
		 command->process == do_update ||
		 command->process == do_rehash ||
		 command->process == do_sendpings ||
		 command->process == do_settings) &&
		 operserv_on == TRUE)
	    {
		notice (s_StatServ, source, ERR_UNKNOWN_COMMAND,
		    strupper (cmd), s_StatServ);
		return;
	    }
	    (*command->process) (source);
	    /* If OperServ is off, we want to snoop several commands. */
	    if ((command->process == do_raw ||
		 command->process == do_inject ||
		 command->process == do_debug ||
		 command->process == do_update ||
		 command->process == do_rehash ||
		 command->process == do_sendpings ||
		 command->process == do_settings) &&
		 operserv_on == FALSE)
	    {
		log ("StatServ: <%s> %s", source, orig);
		snoop (s_StatServ, "<%s> %s", source, orig);
	    }
	}
    }
}

/* Return a help message. */
static void do_help (const char *source) 
{
    char *cmd = strtok (NULL, " ");
            
    if (!cmd)
    {
	notice_list (s_StatServ, source, ss_help_start);
	notice_list (s_StatServ, source, ss_base_help);
	if (tldlookup == TRUE)
	    notice_list (s_StatServ, source, ss_help_tld);
	if (is_oper (source))
	    notice_list (s_StatServ, source, ss_oper_help);
	if (operserv_on == FALSE)
	    notice_list (s_StatServ, source, ss_oper_help_ssonly);
	if (operserv_on == FALSE && is_sra (source))
	    notice_list (s_StatServ, source, ss_sra_help_ssonly);
	if (is_sra (source))
	    notice_list (s_StatServ, source, ss_sra_help);
	notice_list (s_StatServ, source, ss_help_end);
    }
    else
    {
	Hash_HELP *command, hash_table[] =
	{
	    {"STATS",		H_NONE,		stats_help},
	    {"TLD",		H_NONE,		tld_help},
	    {"MAP",		H_IRCOP,	map_help},
	    {"SERVERS",		H_IRCOP,	servers_help},
	    {"AUTH",		H_IRCOP,	auth_help},
	    {"DEAUTH",		H_SRA,		deauth_help},
	    {"FLUSH",		H_SRA,		flush_help},
	    {"RESET",		H_SRA,		reset_help},
	    {"SHUTDOWN",	H_SRA,		shutdown_help},
	    {"RAW",		H_SRA,		raw_help},
	    {"INJECT",		H_SRA,		inject_help},
	    {"DEBUG",		H_SRA,		debug_help},
	    {"UPDATE",		H_SRA,		update_help},
	    {"REHASH",		H_SRA,		rehash_help},
	    {"SENDPINGS",	H_SRA,		sendpings_help},
	    {"SETTINGS",	H_SRA,		settings_help},
	    {NULL}
	};

	if (command = get_help_hash (s_StatServ, source, strupper (cmd),
				     hash_table))
	{
	    if (command->process == tld_help && tldlookup == FALSE)
	    {
		notice (s_StatServ, source, ERR_NO_HELP, "TLD");
		return;
	    }

	    /* These are only available if OperServ is OFF. */
	    if ((command->process == auth_help ||
		 command->process == deauth_help ||
		 command->process == shutdown_help ||
		 command->process == raw_help ||
		 command->process == inject_help ||
		 command->process == debug_help ||
		 command->process == update_help ||
		 command->process == rehash_help ||
		 command->process == sendpings_help ||
		 command->process == settings_help) &&
		 operserv_on == TRUE)
	    {
		notice (s_StatServ, source, ERR_NO_HELP, strupper (cmd));
		return;
	    }

	    notice_list (s_StatServ, source, ss_help_start);
	    if (command->process == stats_help && is_oper (source))
		notice_list (s_StatServ, source, stats_help_oper);
	    else
		notice_list (s_StatServ, source, command->process);
	    notice_list (s_StatServ, source, ss_help_end);
	}
    }
}

/* Show some basic stats. */
static void do_stats (const char *source)
{
    long uptime;
    char *param = strtok (NULL, " ");

    if (!param || param && !is_oper (source))
    {
	uptime = (time (NULL) - me.since);

	notice (s_StatServ, source, SS_STATS_HEADER);
	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_STATS_CURUSER, userstat);
	notice (s_StatServ, source, SS_STATS_MAXAVGUSER, maxusercnt,
	    (userstat + maxusercnt) / 2);
	if (is_oper (source))
	    notice (s_StatServ, source, SS_STATS_CURCHANOPER, countchans (0),
		countchans (1));
	else
	    notice (s_StatServ, source, SS_STATS_CURCHAN, countchans (0));
	notice (s_StatServ, source, SS_STATS_MAXAVGCHAN, maxchancnt,
	    (chanstat + maxchancnt) / 2);
	notice (s_StatServ, source, SS_STATS_CURSERV, servstat);
	notice (s_StatServ, source, SS_STATS_MAXAVGSERV, maxservcnt,
	    (servstat + maxservcnt) / 2);
	if (speedsample > 1000000)
	    notice (s_StatServ, source, SS_STATS_CURSPEEDMB, ((float)
		(speedsample / 1000000)), ((float) (speedstat / 1000000)),
		speedstat > 1000000 ? "MB" : speedstat > 1000 ? "KB" : "bytes");
	else if (speedsample > 1000)
	    notice (s_StatServ, source, SS_STATS_CURSPEEDKB, ((float)
		(speedsample / 1000)), ((float) (speedstat / 1000)),
		speedstat > 1000000 ? "MB" : speedstat > 1000 ? "KB" : "bytes");
	else
	    notice (s_StatServ, source, SS_STATS_CURSPEEDBYTES, speedsample,
		((float) (speedstat / 1000)),
		speedstat > 1000000 ? "MB" : speedstat > 1000 ? "KB" : "bytes");
	notice (s_StatServ, source, SS_STATS_CURTIME, get_time (time (NULL)));
	notice (s_StatServ, source, SS_STATS_MAXUSERTIME, get_time
	    (maxusertime));
	notice (s_StatServ, source, SS_STATS_MAXCHANTIME, get_time
	    (maxchantime));
	notice (s_StatServ, source, SS_STATS_MAXSERVTIME, get_time
	    (maxservtime));
	notice (s_StatServ, source, " ");
	notice (s_StatServ, source, SS_STATS_FOOTER);
    }
    else
    {
	ServStat *serverstat;

	serverstat = findserv (param);

	if (!serverstat)
	{
	    notice (s_StatServ, source, SS_SERVER_NOT_FOUND, param);
	    return;
	}

	notice (s_StatServ, source, SS_SERVSTATS_HEADER, serverstat->name);
	notice (s_StatServ, source, SS_SERVSTATS_SINCE, get_time
	    (serverstat->statssince));
	notice (s_StatServ, source, SS_SERVSTATS_SPLITS, serverstat->splits);
	if (serverstat->is_split == 1)
	{
	    notice (s_StatServ, source, SS_SERVSTATS_DOWNTIME, duration
		(serverstat->splittime + (time (NULL) -
		serverstat->lastsplit)));
	    notice (s_StatServ, source, SS_SERVSTATS_SPLITFOR, time_ago
		(serverstat->lastsplit));
	}
	else
	    notice (s_StatServ, source, SS_SERVSTATS_DOWNTIME, duration
		(serverstat->splittime));
	notice (s_StatServ, source, SS_STATS_FOOTER);
    }
}

/* Show a listing of servers and their version replies. */
static void do_servers (const char *source)
{
    int i;
    ServStat *server;

    notice (s_StatServ, source, SS_SERVERS_HEADER);
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, "%s (operstats-%s. %s%s%s) (0 splits)",
	me.name, version,
	(operserv_on == TRUE) ? "O" : "",
	(statserv_on == TRUE) ? "S" : "",
	debuglevel ? "D" : "");   

    for (i = 0; i < 1024; i++)
    {
	for (server = statlist[i]; server; server = server->next)
	{
	    notice (s_StatServ, source, "%s (%s %s) (%d splits) %s%s%s",
		server->name, server->version, server->flags, server->splits,
		server->is_split ? "(" : "",
		server->is_split ? "Split" : "",
		server->is_split ? ")" : "");
	}
    }
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, SS_SERVERS_FOOTER);
}

/* Display the country/group associated with a given TLD. */
static void do_tld (const char *source)
{
    TLD *tld;
    char *param = strtok (NULL, " ");

    if (tldlookup == FALSE)
    {
	notice (s_StatServ, source, ERR_UNKNOWN_COMMAND, "TLD", s_StatServ);
	return;
    }

    if (!param)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "TLD TopLevelDomain");
	notice (s_StatServ, source, ERR_MORE_INFO, s_StatServ, "TLD");
	return;
    }

    tld = findtld (param);

    if (!tld)
    {
	notice (s_StatServ, source, SS_TLD_NOT_FOUND, param);
	return;
    }
    else
	notice (s_StatServ, source, SS_TLD_IS, tld->name, tld->desc);
}

/* Reset a given server's stats */
static void do_reset (const char *source)
{
    ServStat *server;
    char *param = strtok (NULL, " ");

    if (!param)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "RESET Server");
	notice (s_StatServ, source, ERR_MORE_INFO, s_StatServ, "RESET");
	return;
    }

    server = findserv (param);

    if (!server)
    {
	notice (s_StatServ, source, SS_SERVER_NOT_FOUND, param);
	return;
    }
    else
    {
	server->statssince = time (NULL);
	server->lastsplit = time (NULL);
	server->splittime = 0;
	if (server->is_split)
	    server->splits = 1;
	else
	    server->splits = 0;
	notice (s_StatServ, source, SS_SERVER_RESET, server->name);
    }
}

/* FLUSH: Remove a server from the stats list (ie delinked) */
static void do_flush (const char *source)
{
    ServStat *server;
    char *param = strtok (NULL, " ");

    if (!param)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "FLUSH Server");
	notice (s_StatServ, source, ERR_MORE_INFO, s_StatServ, "FLUSH");
	return;
    }

    server = findserv (param);

    if (!server)
    {
	notice (s_StatServ, source, SS_SERVER_NOT_FOUND, param);
	return;
    }
    else
    {
	int i;

	/* Only flush servers that aren't online. */
	for (i = 0; i < servcnt; i++)
	{
	    if (!stricmp (servlist[i].server, param))
	    {
		notice (s_StatServ, source, SS_CANT_FLUSH);
		return;
	    }
	}
	free (server->version);
	free (server->flags);
	if (server->prev)
	    server->prev->next = server->next;
	else
	    statlist[SHASH(server->name)] = server->next;
	if (server->next)  
	    server->next->prev = server->prev;
	free (server);
	notice (s_StatServ, source, SS_SERVER_FLUSHED, param);
    }
}

/* Show a map of the network. */
static void do_map (const char *source)
{
    int i;
    char l[64];

    /* Find out which server we are connected to */
    for (i = 0; i < servcnt; ++i)
	if (servlist[i].hops == 1)
	    snprintf (l, sizeof (l), "%s", servlist[i].server);

    notice (s_StatServ, source, SS_MAP_HEADER);
    notice (s_StatServ, source, " ");
    showleafs (source, l, "", "");
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, SS_MAP_FOOTER);
}

/* Associated MAP function */
int showleafs (char *source, char *server, char *prevpre, char *arrowchar)
{
    User *user;
    int i = 0, j = 0, k = 0, x, num = 0, cnt;
    char buf[BUFSIZE];
    char pre[BUFSIZE];
    char *nextpre;
    char *s;
    static int depth = 0;
 
    nextpre = malloc (BUFSIZE);
   
    snprintf (pre, sizeof (pre) + 1, prevpre);

    *buf = 0;

    for (i = 0; i < servcnt; i++)
    {
	cnt = 0;

	for (x = 0; x < 1024; x++)
	    for (user = userlist[x]; user; user = user->next)
		if (!strcmp (servlist[i].server, user->server))
		    ++cnt;

	if (servlist[i].hops == 1 && match_wild_nocase (server,
							servlist[i].server))
	{
	    if (servlist[i].lag > 10)
		notice (s_StatServ, source, "%s%s%s (%ds) (%d user%s, %3.2f%%)",
		    "\2", servlist[i].server, "\2", servlist[i].lag, cnt,
		    cnt == 1 ? "" : "s", ((float) (cnt * 100) / usercnt));
	    else
		notice (s_StatServ, source, "%s (%ds) (%d user%s, %3.2f%%)",
		    
		    servlist[i].server, servlist[i].lag, cnt,
		    cnt == 1 ? "" : "s", ((float) (cnt * 100) / usercnt));
	}
	else if (match_wild_nocase (server, servlist[i].server))
	{
	    strcat (buf, pre);
	    strcat (buf, arrowchar);
	    strcat (buf, "-");

	    if (servlist[i].lag > 10)
		strcat (buf, "\2");

	    strcat (buf, servlist[i].server);

	    if (servlist[i].lag > 10)
		notice (s_StatServ, source, "%s%s (%ds) (%d user%s, %3.2f%%)",
		    buf, "\2", servlist[i].lag, cnt, cnt == 1 ? "" : "s",
		    ((float) (cnt * 100) / usercnt));
	    else
		notice (s_StatServ, source, "%s (%ds) (%d user%s, %3.2f%%)",
		    buf, servlist[i].lag, cnt, cnt == 1 ? "" : "s",
		    ((float) (cnt * 100) / usercnt));
	}
    }

    j = numleafs (server);
    k = depth;

    for (i = 0; i < servcnt; i++)
    {
	depth = k;

	if (!stricmp (servlist[i].uplink, server))
	{
	    snprintf (nextpre, strlen (pre) + 1, pre);
	    if (depth++ > 0)
	    {
		if (arrowchar[0] == '`')
		    if (servlist[i].hops == 1)
			strcat (nextpre, "    ");
		    else
			strcat (nextpre, "  ");
		else
		    if (servlist[i].hops == 1)
			strcat (nextpre, "  | ");
		    else
			strcat (nextpre, "| ");
	    }
	    if (j > ++num)
		if (servlist[i].hops == 1)
		    showleafs (source, servlist[i].server, nextpre, "  |");
		else
		    showleafs (source, servlist[i].server, nextpre, "|");
	    else
		if (servlist[i].hops == 1)
		    showleafs (source, servlist[i].server, nextpre, "  `");
		else
		    showleafs (source, servlist[i].server, nextpre, "`");
	}
    }
    free (nextpre);
    depth = 0;
}

/* Associated MAP function */
int numleafs (char *server)
{
    int i, num = 0;
         
    for (i = 0; i < servcnt; ++i)
    {
	if (!stricmp (servlist[i].uplink, server))
	num++;
    }        
    return (num);
}

/* Load the StatServ DB from disk */
void load_ss_dbase (void)
{
    FILE *f = fopen (STATSERV_DB, "r");
    char *item, dBuf[2048], backup[2048];

    if (!f)
    {
	log ("No %s found!", STATSERV_DB);

	/* Check for a db.save */
	strcpy (backup, STATSERV_DB);
	strcat (backup, ".save");
	if (rename (backup, STATSERV_DB) > 0)
	{
	    log ("Recovering %s.save", STATSERV_DB);
	    f = fopen (STATSERV_DB, "r");
	}
	else
	    return;
    }
    
    while (fgets (dBuf, 2047, f))
    {
	item = strtok (dBuf, " ");
	strip (item);

	/* Ignore certain things.. */
	if (*item == '#' || *item == '\n' || *item == '\t' || *item == ' ' ||
	    *item == '\0' || *item == '\r')
	    continue;

	if (!strcmp (item, "ST"))
	{
	    maxusercnt	= atol (strtok (NULL, " "));
	    maxusertime	= atol (strtok (NULL, " "));
	    maxchancnt	= atol (strtok (NULL, " "));
	    maxchantime	= atol (strtok (NULL, " "));
	    maxservcnt	= atol (strtok (NULL, " "));
	    maxservtime	= atol (strtok (NULL, " "));
	    speedstat	= atol (strtok (NULL, " "));
	}

	if (!strcmp (item, "SS"))
	{
	    ServStat *serverstat, **list;

	    serverstat = scalloc (sizeof (ServStat), 1);

	    serverstat->name = sstrdup (strtok (NULL, " "));
	    serverstat->version = sstrdup (strtok (NULL, " "));
	    serverstat->flags = sstrdup (strtok (NULL, " "));
	    serverstat->statssince = atol (strtok (NULL, " "));
	    serverstat->lastsplit = atol (strtok (NULL, " ")); 
	    serverstat->is_split = atol (strtok (NULL, " "));
	    serverstat->splittime = atol (strtok (NULL, " "));
	    serverstat->splits = atol (strtok (NULL, " "));
	    list = &statlist[SHASH(serverstat->name)];
	    serverstat->next = *list;
	    if (*list)
		(*list)->prev = serverstat;
	    *list = serverstat;
	}
    }
    fclose (f);
}

/* Write the stats db to disk. */
void save_ss_dbase ()
{
    FILE *f;
    ServStat *server;
    int i;
    char backup[2048];

    log ("Saving %s...", STATSERV_DB);

    strcpy (backup, STATSERV_DB);  
    strcat (backup, ".save");
    remove (backup);
    if (rename (STATSERV_DB, backup) < 0)
	log ("Can't back up %s!", STATSERV_DB);
    f = fopen (STATSERV_DB, "w");
    
    if (!f)
    {
	log ("Can't write to %s!", STATSERV_DB);
	if (rename (backup, STATSERV_DB) < 0)
	    fatal ("[ERR] Can't restore backup of %s!", STATSERV_DB);
	return;
    }
        
    /* Print an identifying line at the top of the DB. */
    fprintf (f, "OperStats v%s %s\n", version, STATSERV_DB);

    /* Various Statistics */
    fprintf (f, "ST %d %d %d %d %d %d %d \n", maxusercnt, maxusertime,
	maxchancnt, maxchantime, maxservcnt, maxservtime, speedstat);

    /* Per-Server stats */
    for (i = 0; i < 1024; i++)
    {
	for (server = statlist[i]; server; server = server->next)
	{
	    fprintf (f, "SS %s %s %s %d %d %d %d %d \n", server->name,
		server->version, server->flags, server->statssince,
		server->lastsplit, server->is_split, server->splittime,
		server->splits);
	}
    }

    fclose (f);
    remove (backup);
}

/* Add a new TLD into the tld list. */
static TLD * new_tld (const char *name, const char *desc)
{
    TLD *tld;

    if (tld = findtld (name)) /* Don't want dupes, rehash */
    {
	tld->desc = sstrdup (desc);
	return;
    }

    tld = scalloc (sizeof (TLD), 1);

    tld->name = sstrdup (name);
    tld->desc = sstrdup (desc);

    tld->next = tldlist;

    if (tld->next)
	tld->next->prev = tld;

    tldlist = tld;
}

/* Load the TLD listings */
void load_tlds (void)
{
    FILE *f = fopen (TLDFILE, "r");
    char *item, *s, dBuf[BUFSIZE];

    if (!f)
    {
	tldlookup = FALSE;
	return;
    }

    while (fgets (dBuf, 2047, f))
    {
	char *name, *desc;

	s = strtok (dBuf, " ");
	strip (s);

	if (s)
	    name = s;

	s = strtok (NULL, "");
	strip (s);

	if (s)
	    desc = s;

	if (!desc)
	    log ("Found TLD %s but no description!", name);
	else
	    new_tld (name, desc);
    }
    fclose (f);
    tldlookup = TRUE;
}

/* These commands are only usable when OperServ is disabled. */

/* Authenticate for SRA access. */
static void do_auth (const char *source)
{
    char *nick = strtok (NULL, " ");
    char *pass = strtok (NULL, " ");
    User *u;
    SRA *sra;
         
    if (!nick || !pass)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "AUTH NickName Password");
	notice (s_StatServ, source, ERR_MORE_INFO, s_StatServ, "AUTH");
	return;
    }

    u = finduser (source);
    sra = findsra (nick);

    if (!u)
	return;

    if (!sra)
    {
	notice (s_StatServ, source, ERR_NOT_SRA);
	return;
    }

    /* Avoid excessive globals/logging. */
    if (is_sra (source))
    {
	notice (s_StatServ, source, ERR_ALREADY_SRA);
	return;
    }

    if (stricmp (sra->nick, nick) || strcmp (sra->pass, pass))
    {
	notice (s_StatServ, source, ERR_NOT_SRA);
	globops (s_StatServ, OS_FAILED_AUTH, source, u->user, u->host);
	log ("Failed AUTH for %s from %s!%s@%s", sra->nick,
		source, u->user, u->host);
	return;
    }
    else
    {
	u->flags |= UF_SRA;
	globops (s_StatServ, OS_NOW_SRA_GLOBOP, source, u->user, u->host);
	notice (s_StatServ, source, OS_NOW_SRA);
	return;
    }
}

/* De-Authenticate from SRA access. */
static void do_deauth (const char *source)
{
    User *u;

    u = finduser (source);

    if (u->flags & UF_SRA)
    {
	u->flags &= ~UF_SRA;
	notice (s_StatServ, source, OS_NO_LONGER_SRA);
	return;
    }

    /* We shouldn't get here, but just in case. */
    notice (s_StatServ, source, ERR_NOT_SRA);
}

/* Send raw info to the uplink */
static void do_raw (const char *source)
{
    char *text = strtok (NULL, "");

    if (!text)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "RAW Command");
	notice (s_StatServ, source, ERR_MORE_INFO, s_StatServ, "RAW");
    }
    else
	send_cmd (NULL, text);
}

/* Throw something into the inbuf and process it. */
static void do_inject (const char *source)
{
    char *text = strtok (NULL, "");
    char *origbuf = inbuf;

    if (!text)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "INJECT Data");
	notice (s_StatServ, source, ERR_MORE_INFO, s_StatServ, "INJECT");
    }
    else
    {
	strscpy (inbuf, text, BUFSIZE);
	process();
	if (origbuf)
	{
	    /* Something was in the buffer before, process it now. */
	    strscpy (inbuf, origbuf, BUFSIZE);
	    process();
	}
    }
}

/* Toggle debug mode. */
static void do_debug (const char *source)
{
    char *param = strtok (NULL, " ");
        
    if (!param)
    {
	notice (s_StatServ, source, RPL_SYNTAX, "DEBUG Level");
	notice (s_StatServ, source, ERR_MORE_INFO, s_StatServ, "DEBUG");
	return;
    }

    if (atoi (param) > 0)
    {
	debuglevel = atoi (param);
	debug ("-----------------------------------------------");
	notice (s_StatServ, source, OS_DEBUG_LEVEL, atoi (param),
	    debuglevel == 1 ? "Server<->Server traffic" :
	    debuglevel == 2 ? "Some function calls" :
	    debuglevel == 3 ? "Server<->Server traffic and some function calls"
	    : "Unknown");
    }
    else
    {
	notice (s_StatServ, source, OS_DEBUG_LEVEL, atoi (param), "Off");
	debug ("%s turned debug mode off.", source);
	debuglevel = 0;
    }
}

/* Terminate operstats permanently. */
static void do_shutdown (const char *source)
{
    quitmsg = strtok (NULL, "");

    if (!quitmsg)
	quitmsg = "No reason given";

    save_ss_dbase ();

    /* Check for any expired AKills */
    expire_akills ();
    
    if (operserv_on == TRUE)
	save_os_dbase ();

    globops (s_StatServ, OS_SHUTTING_DOWN, source, quitmsg);
    runflags |= RUN_QUITTING;
    runflags |= RUN_TERMINATING;
}

/* Show various OperStats settings */
static void do_settings (const char *source)
{
    notice (s_StatServ, source, OS_SETTINGS_START);
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, OS_SETTINGS_SYNC, update_timeout);
    notice (s_StatServ, source, OS_SETTINGS_PING, ping_delay);
    notice (s_StatServ, source, OS_SETTINGS_CHAN, snoopchan);
    notice (s_StatServ, source, OS_SETTINGS_TLDS, tlds);
    notice (s_StatServ, source, OS_SETTINGS_AKILL, duration (def_akill_time));
    notice (s_StatServ, source, OS_SETTINGS_AKILLREASON, def_akill_reason);
    notice (s_StatServ, source, OS_SETTINGS_AKILLGLOBAL,
	globalonakill_on == TRUE ? "On" : "Off");
    notice (s_StatServ, source, OS_SETTINGS_NONWILD, nonwildreq);
    notice (s_StatServ, source, OS_SETTINGS_DEBUG, debuglevel,
	debuglevel ? "" : " (Off)");
    notice (s_StatServ, source, " ");
    notice (s_StatServ, source, OS_SETTINGS_END);
}

/* Update the DBs. */
static void do_update (const char *source) 
{
    globops (s_StatServ, OS_FORCE_UPDATE, source);
    runflags |= RUN_SAVEDBS;
}

/* Re-read operstats.conf */
static void do_rehash (const char *source) 
{
    /* Turn these off. They'll be re-enabled if they're
       turned on in the conf.
     */
    operserv_on = FALSE;
    statserv_on = FALSE;
    globalnoticer_on = FALSE;
    globalonakill_on = FALSE;
    stamplogons_on = FALSE;

    /* Just in case. */
    do_update (source);

    globops (s_StatServ, OS_DOING_REHASH, CONFIGFILE, source);

    /* Now re-read. */
    load_conf ();
    
    /* Now check it */
    check_conf(); 

    /* Remove any disabled pseudoclients */
    if (ss.isonline && statserv_on == FALSE)
    {
	send_cmd (s_StatServ, "QUIT :Client disabled");
	--userstat;
	ss.isonline = 0;
    }

    if (gn.isonline && globalnoticer_on == FALSE)
    {
	send_cmd (s_GlobalNoticer, "QUIT :Client disabled");
	--userstat;
	gn.isonline = 0;
    }

    /* Re-read TLDs too */
    if (statserv_on == TRUE)
	load_tlds ();
}

/* Ping all servers. */
static void do_sendpings (const char *source)
{
    int i;

    for (i = 0; i < servcnt; ++i)
    {
	send_cmd (me.name, "PING %s %s", me.name, servlist[i].server);
	servlist[i].lastping = time (NULL);
    }
    notice (s_StatServ, source, OS_PINGED_ALL);
}
